/****************************************************************************

 This file is part of the QGLViewer library
 Copyright (C) 2002-2004  Gilles Debunne (Gilles.Debunne@imag.fr)
 Version 1.3.6 Release 1. Packaged on Thursday February 5, 2004.

 http://artis.imag.fr/Members/Gilles.Debunne/QGLViewer

 libQGLViewer is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 libQGLViewer is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with libQGLViewer; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*****************************************************************************/

#include "manipulatedCameraFrame.h"
#include "camera.h"

using namespace std;

/*! Creates a default ManipulatedCameraFrame. Created object is removeFromMouseGrabberPool(). */
ManipulatedCameraFrame::ManipulatedCameraFrame()
  : flyUpVector_(0.0, 1.0, 0.0)
{
  setFlySpeed(0.0);
}


////////////////////////////////////////////////////////////////////////////////

/*! Called for continuous frame motion in fly mode (see ManipulatedFrame::MOVE_FORWARD). Emits
  manipulated(). */
void ManipulatedCameraFrame::updateFrameInFlyMode()
{
  static Vec flyDisp(0.0, 0.0, 0.0);
  switch (action_)
    {
    case MOVE_FORWARD:
      flyDisp.z = -flySpeed();
      translate(inverseTransformOf(flyDisp));
      break;
    case MOVE_BACKWARD:
      flyDisp.z = flySpeed();
      translate(inverseTransformOf(flyDisp));
      break;
    default:
      break;
    }
}

/*! This function will be called by the camera when the camera orientation is changed, so that the
flyUpVector (private) is changed accordingly. You should not need to call this function. */
void ManipulatedCameraFrame::updateFlyUpVector()
{
  flyUpVector_ = inverseTransformOf(Vec(0.0, 1.0, 0.0));
}

////////////////////////////////////////////////////////////////////////////////
//                 M o u s e    h a n d l i n g                               //
////////////////////////////////////////////////////////////////////////////////

// Protected function used to handle mouse events
void ManipulatedCameraFrame::startAction(int ma, bool withConstraint)
{
  ManipulatedFrame::startAction(ma, withConstraint);
  
  switch (action_)
    {
    case MOVE_FORWARD:
    case MOVE_BACKWARD:
      break;
    default:
      break;
    }
}

/*! Move the camera frame according to the mouse motion. The camera is given as an argument so that
its parameters (width and height of the window, fieldOfView) are available in order to fit the mouse
motion and the display. Emits manipulated(). */
void ManipulatedCameraFrame::mouseMoveEvent(QMouseEvent* const e, const Camera* const camera)
{
  // #CONNECTION# QGLViewer::mouseMoveEvent does the updateGL.
  if (action_ != NO_MOUSE_ACTION)
    {
      if (action_ == ZOOM_ON_REGION)
	{
	  prevX_ = e->x();
	  prevY_ = e->y();
	}
      else
	{
	  switch (action_)
	    {
	    case TRANSLATE:
	      {
		Vec trans(static_cast<float>(prevX_ - e->x()), static_cast<float>(e->y() - prevY_), 0.0);
		// Scale to fit the screen mouse displacement
		switch (camera->type())
		  {
		  case Camera::PERSPECTIVE :
		    trans *= 2.0 * tan(camera->fieldOfView()/2.0) *
		      fabs((camera->frame()->coordinatesOf(camera->revolveAroundPoint())).z) / camera->screenHeight();
		    break;
		  case Camera::ORTHO :
		    {
		      double w,h;
		      camera->getOrthoWidthHeight(w, h);
		      trans[0] *= 2.0 * w / camera->screenWidth();
		      trans[1] *= 2.0 * h / camera->screenHeight();
		      break;
		    }
		  }
		translate(inverseTransformOf(translationSensitivity()*trans));
		break;
	      }
	  
	    case MOVE_BACKWARD:
	      {
		Quaternion rot = quaternionFromThetaPhi(e->x(), e->y(), camera);
		rotate(rot);
		translate(inverseTransformOf(Vec(0.0, 0.0, flySpeed())));
		break;
	      }

	    case MOVE_FORWARD:
	      {
		Quaternion rot = quaternionFromThetaPhi(e->x(), e->y(), camera);
		rotate(rot);
		//#CONNECTION# wheelEvent MOVE_FORWARD case
		translate(inverseTransformOf(Vec(0.0, 0.0, -flySpeed())));
		break;
	      }

	    case ZOOM:
	      {
		//#CONNECTION# wheelEvent() ZOOM case
		float coef = max(camera->distanceToSceneCenter(), 0.2f*camera->sceneRadius());
		Vec trans(0.0, 0.0, -coef * (e->y() - prevY_) / camera->screenHeight());
		translate(inverseTransformOf(trans));
		break;
	      }
	  
	    case LOOK_AROUND: 
	      {
		Quaternion rot = quaternionFromThetaPhi(e->x(), e->y(), camera);
		rotate(rot);
		break;
	      }
	  
	    case ROTATE:
	      {
		Vec trans = camera->projectedCoordinatesOf(camera->revolveAroundPoint());
		Quaternion rot = quaternionFromDeformedBall(e->x(), e->y(), trans[0], trans[1], camera);
		rotateAroundPoint(rot, camera->revolveAroundPoint());
		//#CONNECTION# These two functions should go together (spinning detection and activation)
		computeMouseSpeed(e);
		setSpinningQuaternion(rot);
		break;
	      }
	    
	    case SCREEN_ROTATE:
	      {
		Vec trans = camera->projectedCoordinatesOf(camera->revolveAroundPoint());

		const float angle = atan2(e->y() - trans[1], e->x() - trans[0]) - atan2(prevY_-trans[1], prevX_-trans[0]);
	    
		Quaternion rot(Vec(0.0, 0.0, 1.0), angle);
		rotateAroundPoint(rot, camera->revolveAroundPoint());
		updateFlyUpVector();
		//#CONNECTION# These two functions should go together (spinning detection and activation)
		computeMouseSpeed(e);
		setSpinningQuaternion(rot);
		break;
	      }

	    case SCREEN_ROTATE_BOTTOM:
	      {
		const float angle = M_PI * (e->x() - prevX_) / camera->screenWidth();
		Quaternion rot(Vec(0.0, 0.0, 1.0), angle);
		rotate(rot);
		setSpinningQuaternion(rot);
		updateFlyUpVector();
		break;
	      }
	  
	    case SCREEN_TRANSLATE:
	      {
		Vec trans;
		int dir = mouseOriginalDirection(e);
		if (dir == 1)
		  trans.setValue(static_cast<float>(prevX_ - e->x()), 0.0, 0.0);
		else if (dir == -1)
		  trans.setValue(0.0, static_cast<float>(e->y() - prevY_), 0.0);

		switch (camera->type())
		  {
		  case Camera::PERSPECTIVE :
		    trans *= 2.0 * tan(camera->fieldOfView()/2.0) *
		      fabs((camera->frame()->coordinatesOf(camera->revolveAroundPoint())).z) / camera->screenHeight();
		    break;
		  case Camera::ORTHO :
		    {
		      double w,h;
		      camera->getOrthoWidthHeight(w, h);
		      trans[0] *= 2.0 * w / camera->screenWidth();
		      trans[1] *= 2.0 * h / camera->screenHeight();
		      break;
		    }
		  }

		translate(inverseTransformOf(translationSensitivity()*trans));
		break;
	      }

	    case ZOOM_ON_REGION:
	    case NO_MOUSE_ACTION:
	      break;
	    }

	  prevX_ = e->x();
	  prevY_ = e->y();
	}
    }
}

/*! When the user releases the mouse button, the manipulatedFrame action is stopped. */
void ManipulatedCameraFrame::mouseReleaseEvent(QMouseEvent* const e, Camera* const camera)
{
  if (action_ == ZOOM_ON_REGION)
    camera->fitScreenRegion(pressX_, pressY_, e->x(), e->y());
  
  ManipulatedFrame::mouseReleaseEvent(e, camera);
}

////////////////////////////////////////////////////////////////////////////////

/*! Set lastQuat according to the mouse motion. Simple Theta-Phi inferred from the X-flyUpVector() axis. */
Quaternion ManipulatedCameraFrame::quaternionFromThetaPhi(const int x, const int y, const Camera* const camera)
{
  const Quaternion rotX(Vec(1.0, 0.0, 0.0), rotationSensitivity()*static_cast<float>(prevY_-y)/camera->screenHeight());
  const Quaternion rotY(transformOf(flyUpVector()), rotationSensitivity()*static_cast<float>(prevX_-x)/camera->screenWidth());
  return rotY * rotX;
}
